﻿using EmotionsDetector.Helpers;
using FacialTracking;
using Microsoft.ProjectOxford.Emotion;
using Microsoft.ProjectOxford.Emotion.Contract;
using SenseHatDisplay.Helpers;
using SenseHatIO.SenseHatJoystick;
using SenseHatIO.SenseHatLedArray;
using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Threading.Tasks;
using Windows.ApplicationModel;
using Windows.Graphics.Imaging;
using Windows.UI;
using Windows.UI.Popups;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media.Imaging;
using Windows.UI.Xaml.Navigation;

namespace EmotionsDetector
{
    public sealed partial class MainPage : Page, INotifyPropertyChanged
    {
        public event PropertyChangedEventHandler PropertyChanged = delegate { };

        private EmotionServiceClient emotionServiceClient =
            new EmotionServiceClient("YOUR_KEY_GOES_HERE");

        private const string previewStartDescription = "Start preview";
        private const string previewStopDescription = "Stop preview";

        private const string cameraInitilizationError = "Video capture device could not be initialized";
        private const string emotionsApiError = "Emotion API error: ";
            
        private CameraCapture cameraCapture = new CameraCapture();

        private string emotion = string.Empty;

        private WriteableBitmap faceBitmap;

        private double xScalingFactor;
        private double yScalingFactor;

        // IoT-specific components
        private bool isIoTPlatform = false;

        private Joystick joystick;
        private LedArray ledArray;

        private string Emotion
        {
            get { return emotion; }
            set
            {
                emotion = value;
                OnPropertyChanged();
            }
        }

        private WriteableBitmap FaceBitmap
        {
            get { return faceBitmap; }
            set
            {
                faceBitmap = value;
                OnPropertyChanged();
            }
        }

        public MainPage()
        {
            InitializeComponent();

            UpdateUI();
        }

        protected override async void OnNavigatedTo(NavigationEventArgs e)
        {
            base.OnNavigatedTo(e);

            await Initialize();
        }

        private void OnPropertyChanged([CallerMemberName] string propertyName = "")
        {
            PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }

        private async void ButtonPreview_Click(object sender, RoutedEventArgs e)
        {
            await cameraCapture.Initialize(CaptureElementPreview);
            
            if (cameraCapture.IsInitialized)
            {
                await UpdatePreviewState();

                UpdateUI();
            }
            else
            {
                DisplayMessage(cameraInitilizationError);
            }
        }

        private async Task<Emotion> GetEmotion(SoftwareBitmap softwareBitmap)
        {
            Emotion emotion = null;

            try
            {
                var bitmapImageStream = await SoftwareBitmapHelper.GetBitmapStream(softwareBitmap);

                var recognitionResult = await emotionServiceClient.RecognizeAsync(bitmapImageStream);

                emotion = recognitionResult.FirstOrDefault();                
            }
            catch (Exception ex)
            {
                DisplayMessage(emotionsApiError + ex.Message);
            }

            return emotion;
        }

        private void DisplayBitmap(SoftwareBitmap softwareBitmap)
        {
            if (softwareBitmap != null)
            {
                var writeableBitmap = new WriteableBitmap(softwareBitmap.PixelWidth, softwareBitmap.PixelHeight);

                softwareBitmap.CopyToBuffer(writeableBitmap.PixelBuffer);

                FaceBitmap = writeableBitmap;
            }
        }

        private async void ButtonDetectEmotion_Click(object sender, RoutedEventArgs e)
        {
            if (cameraCapture.IsPreviewActive)
            {
                // Capture and display bitmap
                var softwareBitmap = await cameraCapture.CapturePhotoToSoftwareBitmap();
                DisplayBitmap(softwareBitmap);

                // Get and display emotion
                var emotion = await GetEmotion(softwareBitmap);
                DisplayEmotion(softwareBitmap, emotion);
            }
        }

        private void GetScalingFactors(SoftwareBitmap softwareBitmap)
        {
            xScalingFactor = CanvasFaceDisplay.ActualWidth / softwareBitmap.PixelWidth;
            yScalingFactor = CanvasFaceDisplay.ActualHeight / softwareBitmap.PixelHeight;
        }

        private void DisplayEmotion(SoftwareBitmap softwareBitmap, Emotion emotion)
        {
            if (emotion != null)
            {
                var emotionName = EmotionHelper.GetTopEmotionName(emotion);

                DrawFaceBox(softwareBitmap, emotion.FaceRectangle, emotionName);
            }
        }

        private async Task UpdatePreviewState()
        {
            if (!cameraCapture.IsPreviewActive)
            {
                await cameraCapture.Start();                
            }
            else
            {
                await cameraCapture.Stop();

                // Clear UI display
                CanvasFaceDisplay.Children.Clear();
                FaceBitmap = null;
            }
        }

        private void UpdateUI()
        {
            ButtonPreview.Content = cameraCapture.IsPreviewActive ? previewStopDescription : previewStartDescription;
        }

        private async void DisplayMessage(string message)
        {
            if (isIoTPlatform)
            {
                Debug.WriteLine(message);
            }
            else
            {
                var messageDialog = new MessageDialog(message, Package.Current.DisplayName);
                await messageDialog.ShowAsync();
            }
        }

        private void DrawFaceBox(SoftwareBitmap softwareBitmap, Microsoft.ProjectOxford.Common.Rectangle faceRectangle, string emotionName)
        {
            // Clear any previous face rectangles
            CanvasFaceDisplay.Children.Clear();

            // Update scaling factors for displaying face rectangle
            GetScalingFactors(softwareBitmap);

            // Adjust color to emotion
            var emotionColor = EmotionHelper.GetEmotionColor(emotionName);

            // Prepare face box
            var faceBox = EmotionHelper.PrepareFaceBox(faceRectangle, emotionColor, 
                xScalingFactor, yScalingFactor);

            // Prepare emotion description
            var emotionTextBlock = EmotionHelper.PrepareEmotionTextBlock(faceBox, emotionColor, emotionName);

            // Display bounding rectangle, and emotion description
            CanvasFaceDisplay.Children.Add(faceBox);
            CanvasFaceDisplay.Children.Add(emotionTextBlock);
        }

        private async Task Initialize()
        {
            const byte address = 0x46;
            var device = await I2cHelper.GetI2cDevice(address);

            if (device != null)
            {
                joystick = new Joystick(device);
                joystick.ButtonPressed += Joystick_ButtonPressed;

                ledArray = new LedArray(device);

                await cameraCapture.Initialize(CaptureElementPreview);

                isIoTPlatform = true;
            }
        }
        
        private async void Joystick_ButtonPressed(object sender, JoystickEventArgs e)
        {
            if (e.State == JoystickButtonState.Pressed)
            {
                switch (e.Button)
                {
                    case JoystickButton.Up:
                        await UpdatePreviewState();
                        DisplayPreviewStatus();
                        break;

                    case JoystickButton.Enter:
                        await IndicateEmotionOnTheLedArray();
                        break;
                }
            }
        }

        private async Task IndicateEmotionOnTheLedArray()
        {
            if (cameraCapture.IsPreviewActive)
            {
                try
                {
                    // Capture bitmap
                    var softwareBitmap = await cameraCapture.CapturePhotoToSoftwareBitmap();

                    // Get emotion, and its name
                    var emotion = await GetEmotion(softwareBitmap);
                    var emotionName = EmotionHelper.GetTopEmotionName(emotion);

                    // Display emotion color on the LED array
                    var color = EmotionHelper.GetEmotionColor(emotionName);
                    ledArray.Reset(color);
                }
                catch (Exception)
                {
                    Blink(Colors.Red);
                }
            }
            else
            {
                Blink(Colors.Blue);
            }
        }

        private void DisplayPreviewStatus()
        {
            var color = cameraCapture.IsPreviewActive ? Colors.Green : Colors.Blue;

            Blink(color);
        }

        private void Blink(Color color)
        {
            const int msDelayTime = 100;
            const int blinkCount = 2;

            for (int i = 0; i < blinkCount; i++)
            {
                ledArray.Reset(Colors.Black);
                Task.Delay(msDelayTime).Wait();

                ledArray.Reset(color);
                Task.Delay(msDelayTime).Wait();

                ledArray.Reset(Colors.Black);
                Task.Delay(msDelayTime).Wait();
            }
        }
    }
}
